//
// <PW-Id>
//

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright 2018-2020 by Pierre Wolfers Meylan France                       //
//                                                                           //
// This library is free software. Distribution and use rights are outlined   //
// in the file "COPYING" which should have been included with this file.     //
// If this file is missing or damaged, see the license at:                   //
//                                                                           //
//    <To be define when ready>                                              //
//                                                                           //
//   This license described in this file overrides all other licenses that   //
//   might be specified in other files for this library.                     //
//                                                                           //
//   This library is free software; you can redistribute it  and/or modify   //
//   it under the terms of the GNU Lesser General Public License as publi-   //
//   shed by  the  Free Software  Foundation;  either  version 2.1  of the   //
//   License, or (at your option) any later version.                         //
//                                                                           //
//   This library  is  distributed in  the  hope that  it will  be useful,   //
//   but   WITHOUT  ANY  WARRANTY;  without even  the  implied warranty of   //
//   MERCHANTABILITY  or  FITNESS  FOR A  PARTICULAR PURPOSE.  See the GNU   //
//   Library General Public License for more details.                        //
//                                                                           //
//   You should have  received  a copy of  the  GNU Lesser General  Public   //
//   License  along with this library  (see COPYING.LIB); if not, write to   //
//   the Free Software Foundation :                                          //
//                        Inc., 675 Mass Ave, Cambridge, MA 02139, USA.      //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



//
// P.Wolfers DiaViewer Software
//

//
// Module environement to define and manage the Slide object
// by using the image manager for OpenGL ImageMan_REF.
//

#include <setjmp.h>

#include <FL/Fl.H>
#include <FL/fl_ask.H>

#include "DiaViewer_SRV.h"
#include "DiaViewer_DIA.h"
#include "DiaViewer_PRG.h"
#include "DiaViewer_CTX.h"



int IIdeOverflow  =                  0, // Flag for IIde_t overflow.
    SliCacheSize  = DEFAULT_CACHE_SIZE, // Size to use for Image Cache.
    IIdeFld[2]    =  UNIQUE_IDE_FIELDS, // set default Fields.
    IIdeMaxSli    =                  0; // Maximum total number of slide/volume.


static
    IIde_t  IIdeUniqueCnt =  0; // Count to define the Unique integer identifiers.

static int  ImageInitFlag =  1; // Flag for init Cache system.

// Define the structure of Unique identfiers.
static int  IIdeDirClear  =  0; // The slides ide. are unique in volume.
static int  IIdeVolClear  =  1; // The Slides and Dir. are not unique ...
                                // ... for all opened volumes in the run.

// It is implicite that the increment for slide is always 1.
static IIde_t    IIdeDirecIncr, // The increment for directory identifier field.
                 IIdeVolumIncr, // The increment for volume identifier field.
                 IIdeSlideMask, // The bit mask for the slide identifier field.
                 IIdeDirecMask, // The bit mask for the directory identifier field.
                 IIdeVolumMask; // The bit mask for the volume identifier field.


//
// Function to set the IIde increment and mask.
//




static void SetIIdeFields()
// At program init time it is possible to perform this
// function call to choice the IIde_t structure at run time :
//   SetIIdeFields( IIdeFld[0], IIdeFld[1] );
// If Bsli<0, the slide part of Ident is unique for the whole of each Volume.
// If Bdir<0, the directory part of the Ident is unique for all volume.
//
{
    int Bsli = abs( IIdeFld[0] ),
        Bdir = abs( IIdeFld[1] );

    IIdeDirClear = (IIdeFld[0]<0)?-1:0;
    IIdeVolClear = (IIdeFld[1]<0)?-1:0;
    IIdeDirecIncr = 1<<Bsli;
    IIdeVolumIncr = 1<<(Bsli + Bdir);
    IIdeSlideMask = IIdeDirecIncr-1;
    IIdeDirecMask = IIdeVolumIncr-IIdeDirecIncr;
    IIdeVolumMask = ~(IIdeVolumIncr-1);
    IIdeMaxSli    = IIdeVolumIncr-1;
}



//
// Functions for SliEntry class.
//

int  SliEntry::SetSliIIde()
{
    if ((IIdeUniqueCnt&IIdeSlideMask) < IIdeSlideMask) {
        Ident( ++IIdeUniqueCnt );
//printf( " Slide %10x name <%s>\n", Ident(), FName() );
        return 1;
    } else {
        IIdeOverflow    |=   1; // Flag a Slide identifier overflow.
        return 0;
    }
} // SliEntry::SetSliIIde().



SliEntry::SliEntry( const char * fname, IIde_t ide ) :
    Image_REF( fname, ide )
{
    if (ImageInitFlag) { // Force the Cache system init before the first slide allocation.
        SetIIdeFields();
        Image_Init( SliCacheSize );
        ImageInitFlag = 0;
    }
    
    stext_  = NULL;
    owner_  = NULL;
    uflgs_  =    0;
    slid_n_ = film_n_ = para_n_ = chap_n_ = 0;
} // SliEntry::SliEntry( const char * fname, IIde_t ide ).



SliEntry::SliEntry( const uchar * data, int ns, IIde_t ide, char * txt ) :
    Image_REF( data, ide )
{
    if (ImageInitFlag) { // Force the Cache system init before the first slide allocation.
        SetIIdeFields();
        Image_Init( SliCacheSize );
        ImageInitFlag = 0;
    }
    
    owner_  = NULL;
    stext_  = txt;
    film_n_ = para_n_ = chap_n_ = 1; slid_n_ = ns;
} // SliEntry::SliEntry( const uchar * data, IIde_t = 0, int ns, char * txt ).



SliEntry::~SliEntry()
{
    if (stext_) free( stext_ );
} // SliEntry::~SliEntry().


void SliEntry::Rkg_SliProt( int pty, int flg )
//
// For a specified slide, set(flg=1) or Clear(flg=0) ...
// the slide protection bit flags specified by the mask pty.
//
{
    int fla = 0;
    if (pty&1) fla  = Context_ProInb;
    if (pty&2) fla |= Context_ProTxt;
    if (pty&4) fla |= Context_ProOri;
    if (pty&8) fla |= Context_LckRank;

    if (flg) FlgSet( fla );
        else FlgClr( fla );

} // void SliEntry::Rkg_SliProt( int pty, int flg ).





//
// Functions for DirEntry class.
//


int  DirEntry::SetDirIIde()
{
    if ((IIdeUniqueCnt&IIdeDirecMask) < IIdeDirecMask) {
        if (IIdeDirClear) IIdeUniqueCnt &= ~IIdeSlideMask;
        IIdeUniqueCnt +=  IIdeDirecIncr;
        ident_ = IIdeUniqueCnt&~IIdeSlideMask;
//printf( " Directory %10x name <%s>\n", ident_, fname_ );
        return 1;
    } else {
        IIdeOverflow    |=   2; // Flag a Directory identifier overflow.
        return 0;
    }
} // DirEntry::SetDirIIde().



DirEntry::DirEntry( const char *name, IIde_t ide )
{
    fname_ = strdup( name );  // The name is already copied to be keep in sort process.
    stext_ = NULL;
    ident_ =  ide;
    owner_ = NULL;
    dlst_  = NULL;
    slst_  = NULL;
    deep_  =    0;
    udeep_ =    0;
    ndir_  =    0;
    nsli_  =    0;
    cnsli_ =    0;
    flags_ =    0;
} // DirEntry::DirEntry( const char *name, IIde_t ide ).



DirEntry::~DirEntry()
{
    if (ndir_) {
        if (ndir_==1) ((DirEntry*)(dlst_))->~DirEntry();
    } else {
        for(int i=0;i<ndir_;i++) dlst_[i]->~DirEntry();
        delete[] dlst_;
    }
    if (nsli_) {
        if (nsli_==1) ((SliEntry*)(slst_))->~SliEntry();
        else {
            for(int i=0;i<nsli_;i++) slst_[i]->~SliEntry();
            delete[] slst_;
        }
    }
    if (fname_&&deep_)delete[] fname_;
    if (stext_) delete[] stext_;
    owner_ = NULL;
} // DirEntry::~DirEntry().



void DirEntry::SbDir( int nd, DirEntry **tb )
{
    if (nd <= 0) return;
    ndir_ = nd;
    if (nd > 1 ) {
        dlst_ = new DirEntry*[ nd ];
        for (int i=0; i<nd; i++) {
            tb[i]->owner_ = this;
            dlst_[i] = tb[i];
        }
    } else {
        tb[0]->owner_ = this;
        dlst_ = (DirEntry**)tb[0];
    }
    // The delete of table tb must be performed by caller.
} // void DirEntry::SbDir( int nd, DirEntry **tb ).



void DirEntry::Slide( int ns, SliEntry **tb )
{
    if (ns <= 0) return;
    nsli_ = ns;
    if (ns > 1) {
        slst_ = new SliEntry*[ ns ];
        for (int i=0; i<ns; i++) {
            tb[i]->owner_ = this;
            slst_[i] = tb[i];
            tb[i] = NULL;
        }
   } else {
        tb[0]->owner_ = this;
        slst_ = (SliEntry**)tb[0];
        tb[0] = NULL;
    }
    // The delete of table tb must be performed by caller.    
} // void DirEntry::Slide( int ns, SliEntry **tb ).



DirEntry * DirEntry::SbDir( int idx )
{
    if ((ndir_ <= 0)||(idx <= 0)||(idx > ndir_)) return NULL;
    return (ndir_ > 1)? dlst_[idx-1] : (DirEntry*)dlst_;
} // DirEntry * DirEntry::SbDir( int idx = 0 ).



SliEntry * DirEntry::Slide( int idx )
{
    if ((nsli_ <= 0)||(idx <= 0)||(idx > nsli_)) return NULL;
    return (nsli_ > 1)? slst_[idx-1] : (SliEntry*)slst_;    
} // SliEntry * DirEntry::Slide( int idx = 0 ).



void DirEntry::SetDirIdents()
{
    if (SetDirIIde()) {
        if (ndir_ > 0) {
            DirEntry ** tdir = dlst_;
            if (ndir_ > 1 ) {
                for (int ii=0;ii<ndir_;ii++) tdir[ii]->SetDirIdents();
            } else {
                ((DirEntry*)tdir)->SetDirIdents();
            }
        }
        if (nsli_ > 0) {
            SliEntry ** tsli = slst_;
            if (nsli_ > 1) {
                for (int ii=0;ii<nsli_;ii++) tsli[ii]->SetSliIIde();
            } else {
                ((SliEntry*)tsli)->SetSliIIde();
            }
        }
    }
} // DirEntry::SetDirIdents().



SliEntry * DirEntry::SliLocate( const char * fname )
{
    SliEntry * sli = NULL;
    
    if (nsli_) {
        if (nsli_ == 1) {
            sli = (SliEntry*)slst_;
            if (strcmp( fname, sli->FName() ) != 0) sli = NULL;
        } else {
            int ii = 0;
            while (ii < nsli_) {
                sli =slst_[ii++];
                if (strcmp( fname, sli->FName() ) == 0) break;
                sli = NULL;
            }
        }
    }
    return sli;
} // SliEntry * DirEntry::Locate( const char * fname ).





//
// Context input/output Read and Write.
//

#if (defined(WIN32) && !defined(__CYGWIN__)) || defined(__EMX__)
# define sigjmp_buf jmp_buf
# define sigsetjmp( a, b ) setjmp( a )
# define siglongjmp longjmp
#endif


static sigjmp_buf       JmpBuf; // Buffer for long jump on I/O context error. 

static int           VolFlags =    0,
                     VolPrgPc =    0,
                     VolPrgId =    0;

static const char *  VolVname = NULL;

static VarString         FnStr, // Variable Strings used as temporary strings.
                         TxStr;

void OutMemsta( const char * s );       // Routine Reference of GL module to check test for debug.

void SliEntry::SliCntxFlags()
//
// To save a slide context when it was modified.
//
{
    DirEntry * dir = owner_;

//OutMemsta( "SliCntxFlags  1" );
    if (FlgTst( Context_Update )) {
        dir->FlgSet( Context_Update );
        while (dir->LDeep()) {
            dir = dir->Owner();
            dir->FlgSet( Context_SubDir|Context_Update );
        }
        // Now deep is the pointer of Volume root directory.
        VolEntry * vol = (VolEntry*)(dir->Owner());
        vol->FlgSet( Context_Update );
    }
} // void SliEntry::ChkSliCntxFlags().



void SliEntry::RkgClear()
// To Clear the classement data of one slide.
{
    if (!FlgTst( Context_ProInb )) {
        slid_n_ = film_n_ = para_n_ = chap_n_ = 0;
        FlgClr( Context_Slinbr );
        FlgSet( Context_Update );
    }
} // void SliEntry::RkgClear().



static void ErrorCatch( int ierr )
{
    siglongjmp( JmpBuf, ierr );
} // void ErrorCatch( int ierr ).



int  DirEntry::ReadCntxFile( const char * path )
{
    SliEntry                     * sli;
    int     ii = 0,  nsli,  ncv,  flgs;
    char        * fnam,  * fnm,  * txt;
    int icmp, vofl, rdep, cdep, prm[6];

    CntxInsErrCatcher( ErrorCatch );
    if (CntxOpen( path, 0 ) > 0) {      // Open is a Success.
//printf( " ReadCntx : file opened \"%s\"\n", path );  fflush( stdout );
        if (!sigsetjmp( JmpBuf, 1 )) {
//printf( "A md %d, %d\n", CntxHdeIdx, cdeep );  fflush( stdout );
            cdep = LDeep();             // Get the deep of our directory decriptor.
            switch (CntxHdeIdx) {
                // Version 1.000
                case 0:  // Release A00
                case 1:  // Release A01 & A02
                case 2:
                    if (CntxHdeIdx) vofl = CntxGetInt();// Get the volume flag mode.
                               else vofl = (cdep)? 0: 1;
                // We cannot dectect the overlayed volume for old context file versions.
                break;
                // Version 1.001
                case 3:  // Realease A01
                    rdep = CntxGetInt();// Get the deep of directory descriptor.
                    vofl = (cdep == 0);
                default: ;
            }
//printf( "Abis mode %d\n", vofl );  fflush( stdout );
            if (vofl) {                         // Specific to volume root directory.
                // Get and save for VolEntry if we are in Volume root directory.
                VolFlags = CntxGetInt()&~Context_Update;// Get the volume flags and text.
                txt  = CntxGetText(TxStr);      // Get the volume name string.
                VolVname = (txt&&txt[0])? strdup( txt ) : NULL;
//printf( " CntxHdeIdx = %d\n", CntxHdeIdx );
                if (CntxHdeIdx > 2) {           // For complete volume mode (SubVersion > 0).
                    VolPrgPc = CntxGetInt();    // Get Initial position open/select volume.
                    VolPrgId = CntxGetInt();
//printf( " Cntx read RdC Flg %d, Pc %d, Id %d\n", VolFlags, VolPrgPc, VolPrgId );
                }
            }
//printf( "C\n" );  fflush( stdout );
            // Read Directory context setup.
            FlgVal( CntxGetInt()&~Context_Update );     // Set directory flags.
            if (CntxHdeIdx<3) rdep=CntxGetInt();// Set the read deep.
            txt = CntxGetText( TxStr );         // Get and set the directory related text.
            SText( (txt)? txt : NULL );
//printf( "D\n" );  fflush( stdout );
            nsli = NSli();
            // Scan for each slide of this directory.
            if (nsli = NSli()) {                // When we have some slide in this directory.
                do { // We read a slide record.
                    fnm = CntxGetText( FnStr ); // Get the slide filename.
                    ncv =  0;
                    for(int j = 0; j< 6; j++) { // Loop to get orientation, slide #, film #, ...
                        prm[j] = CntxGetInt();
                        if ((j > 1)&&(prm[j])) ncv++;   // ... Chapter #, and volume #.
                    }
                    txt = CntxGetText( TxStr ); // Get the slide anotation text.
                    // We look for a slide that matchs.
                    do {
                        sli = Slide( ++ii );    // Get the current slide descritor ...
                        fnam = sli->FName();    // ... and its filename.
                        icmp = fl_utf_strcasecmp( fnam, fnm );
                    } while ((ii < nsli)&&(icmp < 0));
                    if (icmp == 0) {            // When filenames match ...
                        sli->FlgVal( prm[0]&~Context_Update );  // ... set the slide flags, ...
                        sli->OriOpe( prm[1] );  // ... the Slide original orientation and the ...
                        sli->CurOpe( prm[1] );  // ... ini        tial orientation (Image_REF part).
                        sli->Numb( prm[2] );    // Set the slide #.
                        sli->Film( prm[3] );    // Set the slide film #.
                        sli->Para( prm[4] );    // Set the slide paragraph #.
                        sli->Chap( prm[5] );    // Set the Slide chapter #.
                        sli->Text( (txt)? strdup( txt ):NULL );     // Get the related slide text when present.
                        // Set the flags for present context informations.
                        if (prm[1]) sli->FlgSet( Context_Orient );
                        if (ncv) sli->FlgSet( Context_Slinbr );
                        if (txt) sli->FlgSet( Context_SText );
                    }
                } while (ii < nsli);            // Stop loop on EOF (by use of longjmp) or ...
                                                // ... on end oh SliEntry list.
            }
        }
//printf( "Cntx Close\n" );  fflush( stdout );
        CntxClose();
    } else return 0;

    CntxInsErrCatcher(); 

    // We return 0 when the deep are egal (No error), ...
    // ... a negative value when the current volume include a previous one ...
    // ... and a positive value when the current volume is include in a previous one.
    // The overlay error must be avoid to preserve the context data.
    return rdep - cdep;
} // int ReadCntxFile( const char * fname, int rfl ).



void DirEntry::WriteCntxFile( const char * path )
{
    SliEntry     * sli;
    int deep = LDeep();

//printf( " WriteCntx : deep = %d, path = \"%s\"\n", deep, path );
    if (!FlgTst( Context_Update|Context_SubDir|Context_Data )) return;
    CntxInsErrCatcher( ErrorCatch );
    if (CntxOpen( path, 1 ) > 0) {              // Open to write Success.
        if (!sigsetjmp( JmpBuf, 1 )) {
            CntxPutInt( deep, 2 );              // Write the current directory deep.
            CntxPutEoln();
            if (deep == 0) {
                // Put for VolEntry if we are in Volume root directory.
//printf(" WrtCntx : A VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
                CntxPutInt( VolFlags, 6 );      // Put Volume flags.
                CntxPutText( VolVname );        // Put the Volume name (do not delete!).
//printf(" WrtCntx : B VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
                VolVname = NULL;
                CntxPutInt( VolPrgPc, 6 );      // Write the last position pc and index.
                CntxPutInt( VolPrgId, 6 );
                CntxPutEoln();
            }
            // Put the directory flags and attached text.
            CntxPutInt( FlgVal() & ~Context_Update, 6 );
            CntxPutText( SText() );
            // Create each slide context when required.
            for(int i = 1; i <= NSli(); i++) {
                sli = Slide( i );
                if (sli->FlgTst( Context_Data )) { // Some data must be written (or rewritten).
                    CntxPutText( sli->FName() );
                    CntxPutInt( sli->FlgVal() & ~Context_Update, 6 );
                    CntxPutInt( sli->CurOpe(), 2  );
                    CntxPutInt( sli->Numb(), 6 ); CntxPutInt( sli->Film(), 6 );
                    CntxPutInt( sli->Para(), 4 ); CntxPutInt( sli->Chap(), 4 );
                    CntxPutText( sli->Text() );
                }
            }
        }
        CntxClose();
    }
    CntxInsErrCatcher(); 
} // int WriteCntxFile( const char * path ).



void DirEntry::WriteDirCntx()
{
    int nd = NDir();
///     ns = NSli();
    DirEntry  * dir;

    if  (FlgTst( Context_Update|Context_SubDir )) {
        // If somethings changed in the slide volume. 
        if (nd) { // Loop for recursive call on all slide sub-directories.
            for(int i=1;i<=nd;i++) {
                dir = SbDir( i );                       // For each sub-directory.
                if (dir->FlgTst( Context_Update|Context_SubDir )) {
                    // If the directory (or su-directories) context was updated.
                    CPath->Set( dir->FName(), 1 );      // Set the Directory path.
                    dir->WriteDirCntx();                // Create the local context file.
                    CPath->Rem();                       // Restore previous context.
                }
            }
        }
        if (/*ns&&*/FlgTst( Context_Update|Context_SubDir )) {
            CPath->Set( CNTX_FNAME );
            WriteCntxFile( CPath->Path() );        // Write context for the current directory.
            CPath->Rem();
        }
    }
} // int DirEntry::WriteDirCntx().




void DirEntry::RkgClear()
// Clear classement data of all slides of a current directory.
{
    int ns = NSli();

    for (int i = 1; i <= ns; i++) Slide( i )->RkgClear();
    FlgClr( Context_Slinbr );
    FlgSet( Context_Update );
} // DirEntry::RkgClear().


/*
void SliEntry::InitBltPath( VolEntry * vol )
// Init the Reverse Path system at the Root ofslide volume.
//
{   // Create and Initialize the path slide access to the volume directory.
    CPath = new FilePath( vol->TDeep() + 1 );
//  CPath->Set( rootd_->FName(), 1 );   // ... Set the volume root directory path.
} // void VolEntry::InitBltPath().



void SliEntry::ReleaseBltPath()
{
    delete CPath;
    CPath = NULL;
} // VolEntry::ReleaseBltPath().
*/


void DirEntry::BltPath()
// Sub-Routine of the function SliEntry::BltPAth().
{
    if (deep_) {
        owner_->BltPath();      // Recursive call to find the volume root.
        CPath->Set( FName(), 1 );
    } else {
        CPath->Clear();         // For Root, erase and put root filename to begin the Path.
        CPath->Set( FName(), 1 );
    }
} // void DirEntry::BltPath().



const char * SliEntry::BltPath()
// Routine to call to get the filepath of the current slide ...
// ... without follow the way from the root directory of the volume ...
// ... allow a random access at any slide of the source volume.
{
    owner_->BltPath();          // Call to build directories path make, ...
    CPath->Set( FName(), 0 );   // ... append the slide filename and ...
    return  CPath->Path();      // Return the complete slide file path.
} // char * SliEntry::BltPath().



//
// Functions for VolEntry class.
//


int  VolEntry::SetVolIIde()
{
    if ((IIdeUniqueCnt&IIdeVolumMask) < IIdeVolumMask) {
        if (IIdeVolClear) IIdeUniqueCnt &= IIdeVolumMask;
        IIdeUniqueCnt +=  IIdeVolumIncr;
        ident_ = IIdeUniqueCnt&IIdeVolumMask;
//printf( " Volume %10x name <%s>\n", ident_, fname_ );
        return 1;
    } else {
        IIdeOverflow    |=   4; // Flag a Volume identifier overflow.
        return 0;
    }
} // VolEntry::SetVolIIde().



VolEntry::VolEntry( const char *name, DirEntry * root , int idx )
{
    vname_ = NULL;
    rootd_ = root;                      // Set the link to the root directory.
    root->Owner( (DirEntry*)this );     // Set the back link to the Volume record. 
    scprg_ = NULL;
    ident_ =    0;
    tdeep_ =    0;
    nwind_ =    0;
    tndir_ =    0;
    tnsli_ =    0;
    index_ =  idx;
    flags_ =    0;
    prcnt_ =    0;
    pridx_ =    0;
    volfl_ =    0;

    OpenVolTable.AddObj( (void*)this ); // Append the new volume to the open volume table.

} // VolEntry::VolEntry( const char *name, DirEntry * root = NULL, int idx = 0 ).



VolEntry::~VolEntry()
{
    OpenVolTable.RemObj( (void*)this ); // Remove the volume of the open volume table.

    if (rootd_) delete rootd_;          // Free all the volume tree.
    if (vname_) delete[] vname_;
//  if (stext_) delete[] stext_;        // No specific text for volume (use Dir. text).
} // VolEntry::~VolEntry().



void VolEntry::VName( const char * name )
{
    if (vname_) delete[] vname_;
    if (name&&name[0]) vname_ = strdup( name );
                  else vname_ = NULL;
    FlgSet( Context_Update );
} // void VolEntry::VName( const char * name ).



void VolEntry::SetVolContext()
//
// To use when the volume structure is build after the root directory.
//
{
    FlgVal( VolFlags );
//printf( "SetVolCntx :  Set VName at \"%s\" for \"%s\" =>\n", VolVname, rootd_->FName() );
    VName( VolVname  );
    VolVname    = NULL;
    prcnt_  = VolPrgPc;
    pridx_  = VolPrgId;
} // VolEntry::SetVolContext().




void VolEntry::WriteVolContext()
{
//printf( " WVol : A\n" ); fflush( stdout );
    if (FlgTst( Context_Update )) {
        DirEntry * root = VRoot();
        VolFlags =  FlgVal()&Context_Viewer;
        VolVname =  VName();
//printf(" WVol : B VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
        VolPrgPc =  PrgPc();
        VolPrgId =  PrgId();
//printf(" WVol : C VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
        CPath = new FilePath( TDeep() + 1 );
        CPath->Set( root->FName(), 1 );
//printf(" WVol : D VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
        root->WriteDirCntx();
//printf(" WVol : E VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
        delete CPath;
//printf(" WVol : F VName = \"%s\"\n", (VolVname)? VolVname : "<>" ); fflush( stdout );
        VolFlags =    0;
        VolVname = NULL;
        VolPrgPc =    0;
        VolPrgId =    0;
    }
} // void VolEntry::WriteVolContexts().



void VolEntry::SetVolIdents()
//
// To set the various identifier in a slide volume.
//
{
    this->SetVolIIde();         // Set a volume identifier.                   
    rootd_->SetDirIdents();     // Set identifiers for each directory and slide of volume.
    rootd_->Owner( (DirEntry*)this );
} // VolEntry::SetVolIdents().






//
// end of <PW-Id>.
//
